"""
Simulation data recorder.

Maxwell X. Cai, April 2021.

"""
import numpy as np 


class SimulationDataRecorder(object):

    def __init__(self, particles=None, quantities=['a', 'ecc', 'inc'], buffer_len=0):

        self._monitored_particles = particles
        self._monitored_quantities = quantities
        self._recording = False 
        self._particles_hash_vec = None 

        # initialize storage
        self.data = dict()

    def reset(self):
        del self.data
        self.data = dict()
    
    def start(self):
        self._recording = True 

    def stop(self):
        self._recording = False

    def set_monitored_quantities(self, quantities):
        self._monitored_quantities = quantities

    def get_monitored_quantities(self, quantities):
        return self._monitored_quantities

    def set_monitored_particles(self, particles):
        self._monitored_particles = particles

    def get_monitored_particles(self, particles):
        self._monitored_particles = particles
 
    def record(self, state_dict):
        if self._recording is False:
            return 

        if self._monitored_quantities is None:
            quantities = ['a', 'ecc', 'inc', 'x', 'y', 'z', 'vx', 'vy', 'vz', 'mass', 'time']
        else:
            quantities = self._monitored_quantities

        for i, qty in enumerate(quantities):
            if qty in state_dict:
                d = state_dict[qty]
                if d.ndim == 2:
                    d = d[~np.isnan(d).any(axis=1)] # remove the parts of buffer that contains NaN
                else:
                    d = d[~np.isnan(d)]
                
                if qty in self.data and self.data[qty] is not None:
                    # append
                    indices = np.where(np.in1d(self._particles_hash_vec, state_dict['hash'][0]))[0] # the indices to be updated
                    if self._monitored_particles is not None:
                        if d.ndim == 2:
                            tmp_data = np.zeros((d.shape[0], self.data[qty].shape[1])) * np.nan 
                            tmp_data[:, indices] = d[self._monitored_particles]
                            # print(tmp_data, indices, self._particles_hash_vec, state_dict['hash'][0])
                            self.data[qty] = np.append(self.data[qty], tmp_data, axis=0)
                        else:
                            self.data[qty] = np.append(self.data[qty], d[self._monitored_particles])
                    else:
                        if d.ndim == 2:
                            tmp_data = np.zeros((d.shape[0], self.data[qty].shape[1])) * np.nan 
                            tmp_data[:, indices] = d
                            # print(tmp_data, indices, self._particles_hash_vec, state_dict['hash'][0])
                            self.data[qty] = np.append(self.data[qty], tmp_data, axis=0)
                        else:
                            self.data[qty] = np.append(self.data[qty], d)
                else:
                    # New record
                    self._particles_hash_vec = state_dict['hash'][0] # get the hash of all particles at t = 0
                    if self._monitored_particles is not None:
                        self.data[qty] = d[self._monitored_particles].copy()
                    else:
                        self.data[qty] = d.copy()




from setuptools import setup, find_packages
try:
    from setuptools import setup, Extension
except ImportError:
    from distutils.core import setup, Extension

from distutils.command.build_ext import build_ext
import sys
import sysconfig
import os



def locate_cuda():
    if 'CUDAHOME' in os.environ:
        home = os.environ['CUDAHOME']
        nvcc = os.path.join(home, 'bin', 'nvcc')
    else:
        for directory in os.environ['PATH'].split(os.pathsep):
            nvcc = os.path.join(directory, 'nvcc')
            if os.path.exists(nvcc):
                # further check whether the header files and library are there
                home = os.path.dirname(os.path.dirname(nvcc))
                if os.path.exists(os.path.join(home, 'include')) and os.path.exists(os.path.join(home, 'lib64')):
                    cudaconfig = {'home': home, 'nvcc': nvcc, 'include': os.path.join(home, 'include'), 'lib64': os.path.join(home, 'lib64')}
                    return cudaconfig
        return None

def customize_compiler_for_nvcc(self):
    """
    inject deep into distutils to customize how the dispatch
    to gcc/nvcc works.
    If you subclass UnixCCompiler, it's not trivial to get your subclass
    injected in, and still have the right customizations (i.e.
    distutils.sysconfig.customize_compiler) run on it. So instead of going
    the OO route, I have this. Note, it's kindof like a wierd functional
    subclassing going on.
    """

    # tell the compiler it can processes .cu
    self.src_extensions.append('.cu')

    # save references to the default compiler_so and _comple methods
    default_compiler_so = self.compiler_so
    super = self._compile

    # now redefine the _compile method. This gets executed for each
    # object but distutils doesn't have the ability to change compilers
    # based on source extension: we add it.
    def _compile(obj, src, ext, cc_args, extra_postargs, pp_opts):
        if os.path.splitext(src)[1] == '.cu':
            # use the cuda for .cu files
            self.set_executable('compiler_so', CUDA['nvcc'])
            # use only a subset of the extra_postargs, which are 1-1 translated
            # from the extra_compile_args in the Extension class
            postargs = extra_postargs['nvcc']
        else:
            postargs = extra_postargs['gcc']

        super(obj, src, ext, cc_args, postargs, pp_opts)
        # reset the default compiler_so, which we might have changed for cuda
        self.compiler_so = default_compiler_so

    # inject our redefined _compile method into the class
    self._compile = _compile

# run the customize_compiler
class cuda_build_ext(build_ext):
    def build_extensions(self):
        customize_compiler_for_nvcc(self.compiler)
        build_ext.build_extensions(self)

def find_files_with_ext(path, ext):
    files_list = list()
    for root, dirs, files in os.walk(path):
        for name in files:
            try:
                current_ext = name.split(".")[-1]
            except IndexError:
                continue

            if current_ext == ext:
                files_list.append(join(root, name))
    return files_list

suffix = sysconfig.get_config_var('EXT_SUFFIX')
if suffix is None:
    suffix = ".so"

extra_link_args = []
if sys.platform == 'darwin':
    from distutils import sysconfig
    vars = sysconfig.get_config_vars()
    vars['LDSHARED'] = vars['LDSHARED'].replace('-bundle', '-shared')
    extra_link_args=['-Wl,-install_name,@rpath/libabie'+suffix]

sources = ['src/integrator_gauss_radau15.c',
           'src/integrator_wisdom_holman.c',
           'src/integrator_runge_kutta.c',
           'src/common.c',
           'src/additional_forces.c']

# Locate CUDA paths
CUDA = locate_cuda()

gcc_compile_flags = ['-fstrict-aliasing', '-O3','-std=c99','-fPIC', '-shared', '-fcommon', '-fcommon']
if CUDA is not None:
    # CUDA/GPU detected on the system
    sources.append('src/gpuforce.cu')
    sources.append('src/bodysystemcuda.cu')
    extra_link_args = ['-fPIC', '-DGPU', '-O3', '-g', '-lcudart', '-lstdc++']
    gcc_compile_flags.append('-DGPU')
    print('CUDA SDK found! ABIE will be built with GPU support.')
else:
    print('CUDA SDK  not found! ABIE will be built without GPU support.')

module_abie = Extension(name = 'libabie',
                        sources = sources,
                        extra_link_args=extra_link_args,
                        extra_compile_args={
                            'gcc': gcc_compile_flags,
                            'nvcc': ['-O3', '-Xcompiler', '-DGPU', '--use_fast_math', '--ptxas-options=-v', '-c', '--compiler-options', '-fPIC']
                            },
                        )

setup(name='abie',
      version='0.5.2',
      description='Alice-Bob Integrator Environment (ABIE), a GPU-accelerated integrator framework for astrophysical N-body simulations',
      url='https://github.com/MovingPlanetsAround/ABIE',
      author='Maxwell X. Cai, Javier Roa, Adrian S. Hamers, Nathan W. C. Leigh',
      author_email='maxwellemail@gmail.com',
      license='BSD 2-Clause',
      packages=find_packages(),
      zip_safe=False,
      install_requires=['toml', 'numpy', 'h5py'],
      entry_points={'console_scripts': ['abie = ABIE.abie:main'] },
      cmdclass={'build_ext': cuda_build_ext,},
      ext_modules = [module_abie],
      )
